from cryptography.hazmat.primitives.asymmetric import rsa, padding
from cryptography.hazmat.primitives.ciphers import Cipher, algorithms, modes
from cryptography.hazmat.primitives import hashes, serialization
from cryptography.hazmat.backends import default_backend
from concurrent.futures import ThreadPoolExecutor
import base64
import os
import hashlib

def ParseHash(HashName: str) -> hashes.HashAlgorithm | None:
    if (HashName == "sha224"):
        return hashes.SHA224()
    elif (HashName == "sha256"):
        return hashes.SHA256()
    elif (HashName == "sha384"):
        return hashes.SHA384()
    elif (HashName == "sha512"):
        return hashes.SHA512()
    elif (HashName == "sha1"):
        return hashes.SHA1()
    elif (HashName == "none"):
        return None
    
    raise ValueError("Invalid hash name.")

def GenerateRSAKeys(Size: int = 8192) -> tuple[rsa.RSAPrivateKey, rsa.RSAPublicKey]:
    privateKey = rsa.generate_private_key(65537, Size)
    publicKey = privateKey.public_key()

    return privateKey, publicKey

def SaveKeys(
    PrivateKey: rsa.RSAPrivateKey | None,
    PrivateFile: str | None,
    PrivatePassword: str,
    PublicKey: rsa.RSAPublicKey | None,
    PublicFile: str | None
) -> tuple[bytes | None, bytes | None]:
    privatePem = PrivateKey.private_bytes(
        encoding = serialization.Encoding.PEM,
        format = serialization.PrivateFormat.PKCS8,
        encryption_algorithm = serialization.BestAvailableEncryption(PrivatePassword.encode("utf-8"))
    ) if (PrivateKey is not None) else None
    publicPem = PublicKey.public_bytes(
        encoding = serialization.Encoding.PEM,
        format = serialization.PublicFormat.SubjectPublicKeyInfo
    ) if (PublicKey is not None) else None

    if (PrivateKey is not None and PrivateFile is not None):
        with open(PrivateFile, "wb") as f:
            f.write(privatePem)
    
    if (PublicKey is not None and PublicFile is not None):
        with open(PublicFile, "wb") as f:
            f.write(publicPem)
    
    return (
        base64.b64encode(privatePem) if (privatePem is not None) else None,
        base64.b64encode(publicPem) if (publicPem is not None) else None
    )

def LoadKeysFromFile(
    PrivateFile: str,
    PrivatePassword: str,
    PublicFile: str
) -> tuple[rsa.RSAPrivateKey, rsa.RSAPublicKey]:
    with open(PrivateFile, "rb") as f:
        privatePem = f.read()
    
    with open(PublicFile, "rb") as f:
        publicPem = f.read()
    
    return LoadKeysFromContent(privatePem, PrivatePassword, publicPem)

def LoadKeysFromContent(
    PrivateContent: str | bytes | None,
    PrivatePassword: str,
    PublicContent: str | bytes | None
) -> tuple[rsa.RSAPrivateKey | None, rsa.RSAPublicKey | None]:
    if (PrivateContent is None):
        privateKey = None
    else:
        privateKey = serialization.load_pem_private_key(
            data = PrivateContent.encode("utf-8") if (isinstance(PrivateContent, str)) else PrivateContent,
            password = PrivatePassword.encode("utf-8"),
            backend = default_backend()
        )
    
    if (PublicContent is None):
        publicKey = None
    else:
        publicKey = serialization.load_pem_public_key(
            data = PublicContent.encode("utf-8") if (isinstance(PublicContent, str)) else PublicContent,
            backend = default_backend()
        )

    return (privateKey, publicKey)

def Encrypt(Hash: hashes.HashAlgorithm | None, PublicKey: rsa.RSAPublicKey, Data: str | bytes, MaxThreads: int) -> str | bytes:
    def _encrypt(Key: bytes, Nonce: bytes, Idx: int, Chunk: bytes, ChunkSize: int) -> tuple[int, bytes]:
        nonceInt = int.from_bytes(Nonce, "big")
        blocksPerChunk = ChunkSize // 16
        offset = Idx * blocksPerChunk
        derivedNonce = (nonceInt + offset).to_bytes(16, "big")

        cipher = Cipher(algorithms.AES(Key), modes.CTR(derivedNonce), backend = default_backend())
        encryptor = cipher.encryptor()
        
        return (Idx, encryptor.update(Chunk) + encryptor.finalize())

    if (Hash is None):
        return Data
    
    if (isinstance(Data, bytes)):
        returnAsBytes = True
        data = Data
    else:
        returnAsBytes = False
        data = Data.encode("utf-8")
    
    key = os.urandom(32)
    encriptedKey = PublicKey.encrypt(
        key,
        padding.OAEP(
            mgf = padding.MGF1(Hash),
            algorithm = Hash,
            label = None
        )
    )

    nonce = os.urandom(16)
    chunkSize = 1024 * 1024
    chunks = [data[i:i + chunkSize] for i in range(0, len(data), chunkSize)]

    results = [None] * len(chunks)

    with ThreadPoolExecutor(max_workers = MaxThreads) as executor:
        futures = [
            executor.submit(_encrypt, key, nonce, idx, chunk, chunkSize)
            for idx, chunk in enumerate(chunks)
        ]

        for future in futures:
            idx, ciphertext = future.result()
            results[idx] = ciphertext

    result = (
        len(encriptedKey).to_bytes(4, "big") +
        encriptedKey +
        nonce +
        b"".join(results)
    )
    result = base64.b64encode(result)

    if (returnAsBytes):
        return result
    
    return result.decode("utf-8")

def Decrypt(Hash: hashes.HashAlgorithm | None, PrivateKey: rsa.RSAPrivateKey, Data: str | bytes, MaxThreads: int) -> str | bytes:
    def _decrypt(Key: bytes, Nonce: bytes, Idx: int, Chunk: bytes, ChunkSize: int) -> tuple[int, bytes]:
        nonceInt = int.from_bytes(Nonce, "big")
        blocksPerChunks = ChunkSize // 16
        offset = Idx * blocksPerChunks

        newNonce = (nonceInt + offset).to_bytes(16, "big")
        cipher = Cipher(algorithms.AES(Key), modes.CTR(newNonce), backend = default_backend())
        decryptor = cipher.decryptor()

        return (Idx, decryptor.update(Chunk) + decryptor.finalize())
    
    if (Hash is None):
        return Data

    if (isinstance(Data, bytes)):
        returnAsBytes = True
        data = Data
    else:
        returnAsBytes = False
        data = Data.encode("utf-8")
    
    data = base64.b64decode(data)
    
    lenKey = int.from_bytes(data[:4], "big")
    encryptedKey = data[4:4 + lenKey]
    nonce = data[4 + lenKey:4 + lenKey + 16]
    ciphertext = data[4 + lenKey + 16:]

    key = PrivateKey.decrypt(
        encryptedKey,
        padding.OAEP(
            mgf = padding.MGF1(Hash),
            algorithm = Hash,
            label = None
        )
    )
    chunkSize = 1024 * 1024
    chunks = [
        ciphertext[i:i + chunkSize]
        for i in range(0, len(ciphertext), chunkSize)
    ]

    results = [None] * len(chunks)

    with ThreadPoolExecutor(max_workers = MaxThreads) as executor:
        futures = [
            executor.submit(_decrypt, key, nonce, idx, chunk, chunkSize)
            for idx, chunk in enumerate(chunks)
        ]

        for future in futures:
            idx, plaintext = future.result()
            results[idx] = plaintext

    if (returnAsBytes):
        return b"".join(results)
    
    return "".join([p.decode("utf-8") for p in results])

def HashContent(Content: str | bytes, Hash: hashes.HashAlgorithm) -> str:
    if (Hash is None):
        raise TypeError("Hash is None. None type is not valid for hashing content.")
    
    content = Content.encode("utf-8") if (isinstance(Content, str)) else Content
    
    if (isinstance(Hash, hashes.SHA224)):
        hashObj = hashlib.sha224(content)
    elif (isinstance(Hash, hashes.SHA256)):
        hashObj = hashlib.sha256(content)
    elif (isinstance(Hash, hashes.SHA384)):
        hashObj = hashlib.sha384(content)
    elif (isinstance(Hash, hashes.SHA512)):
        hashObj = hashlib.sha512(content)
    elif (isinstance(Hash, hashes.SHA1)):
        hashObj = hashlib.sha1(content)
    
    hashHex = hashObj.hexdigest()
    
    return hashHex


from typing import Any, Self

class ClientConfiguration():
    def __init__(self) -> None:
        # Encryption configuration
        self.Encryption_PublicKey: Any | None = None
        self.Encryption_PrivateKey: Any | None = None
        self.Encryption_PrivateKeyPassword: str = "changeme"
        self.Encryption_Threads: int = 1
        self.Encryption_RSASize: int = 4096
        self.Encryption_Hash: str = "sha512"

        # Service configuration
        self.Service_DefaultAPIKey: str = "nokey"

        # Other configuration
        self.PingInterval: float = 2

    def ToDict(self, SavePublicKey: bool = False) -> dict[str, Any]:
        d = self.__dict__.copy()

        if (not SavePublicKey):
            d["Encryption_PublicKey"] = None
            d["Encryption_PrivateKey"] = None
        
        return d
    
    @classmethod
    def FromDict(cls, Dict = dict[str, Any]) -> Self:
        instance = cls.__new__(cls)

        for k, v in Dict.items():
            setattr(instance, k, v)
        
        return instance


from html2text import HTML2Text

def HTML_To_Markdown(Content: str) -> str:
    h = HTML2Text(bodywidth = 0)
    h.single_line_break = True

    content = h.handle(Content)
    
    h.close()
    return content


from typing import Any, Literal
from collections.abc import AsyncGenerator
from websockets import connect as WS_Connect
from websockets.protocol import State as WS_State
from . import configuration, encryption
import json
import base64
import asyncio

VERSION: int = 180000
TRANSFER_RATE = 8192 * 1024

class ClientSocket():
    def __init__(
        self,
        Type: Literal["websocket"],
        Configuration: configuration.ClientConfiguration
    ) -> None:
        self.__socket__ = None
        self.__socket_type__ = Type  # TODO: More socket types in the future!
        self.__configuration__ = Configuration
        self.__server_public_key_str__ = None
        self.__server_public_key__ = None
        self.__current_connection__ = [None, None, None, None]

        if (Configuration.Encryption_PublicKey is None or Configuration.Encryption_PrivateKey is None):
            self.__private_key__, self.__public_key__ = encryption.GenerateRSAKeys(Size = Configuration.Encryption_RSASize)
        
        _, self.__public_key_str__ = encryption.SaveKeys(None, None, "", self.__public_key__, None)
        self.__public_key_str__ = self.__public_key_str__.decode("utf-8")
    
    def IsConnected(self) -> bool:
        if (self.__socket__ is None):
            return False  # Make sure the socket is not null
        
        if (self.__socket_type__ == "websocket"):
            return self.__socket__.state == WS_State.OPEN  # Make sure the connection is OPEN
        
        return False

    async def Connect(self, Host: str, Port: int, Secure: bool = False) -> None:
        await self.Close()

        if (self.__socket_type__ == "websocket"):
            uri = ("wss" if (Secure) else "ws") + f"://{Host}:{Port}"
            self.__socket__ = await WS_Connect(
                uri = uri,
                max_size = TRANSFER_RATE,
                ping_interval = self.__configuration__.PingInterval,
                ping_timeout = None
            )
        
            while (self.__socket__.state == WS_State.CONNECTING):
                await asyncio.sleep(0.1)
        
        self.__current_connection__ = [self.__socket_type__, Host, Port, Secure]
        await self.__set_server_public_key__()

    async def Close(self) -> None:
        if (self.__socket__ is None):
            return
        
        if (self.__socket_type__ == "websocket"):
            try:
                await self.Send("close")
                await self.__socket__.close()
            except:
                pass
        
        self.__socket__ = None
    
    async def __set_server_public_key__(self) -> None:
        self.__server_public_key_str__ = await self.SendAndReceive("get_public_key")
        _, self.__server_public_key__ = encryption.LoadKeysFromContent(None, "", base64.b64decode(self.__server_public_key_str__))

    async def __send__(self, Data: str) -> None:
        if (not self.IsConnected()):
            raise ConnectionError("Socket not connected.")

        if (self.__socket_type__ == "websocket"):
            await self.__socket__.send(Data)
    
    async def __recv__(self) -> str:
        if (not self.IsConnected()):
            raise ConnectionError("Socket not connected.")
        
        if (self.__socket_type__ == "websocket"):
            recv = await self.__socket__.recv(decode = True)
        
        return recv
    
    async def Send(self, Data: str) -> None:
        chunks = [Data[i:i + TRANSFER_RATE] for i in range(0, len(Data), TRANSFER_RATE)]

        for chunk in chunks:
            await self.__send__(chunk)
        
        await self.__send__("--END--")
    
    async def Receive(self) -> str:
        data = ""

        while (True):
            chunk = await self.__recv__()

            if (chunk == "--END--"):
                break
            
            chunk = chunk[:TRANSFER_RATE]
            data += chunk
        
        return data

    async def SendAndReceive(self, Data: str) -> str:
        await self.Send(Data)
        return await self.Receive()

    async def AdvancedSendAndReceive(
        self,
        ModelName: str,
        Key: str | None = None,
        PromptConversation: list[dict[str, str | dict[str, str | bytes]]] = [],
        PromptParameters: dict[str, Any] = {},
        UserParameters: dict[str, Any] = {},
        Service: str = "inference"
    ) -> AsyncGenerator[dict[str, Any]]:
        if (self.__server_public_key__ is None):
            await self.__set_server_public_key__()

        h = encryption.ParseHash(self.__configuration__.Encryption_Hash)
        data = {
            "hash": self.__configuration__.Encryption_Hash,
            "public_key": self.__public_key_str__,
            "version": VERSION,
            "content": {
                "model_name": ModelName,
                "service": Service,
                "key": self.__configuration__.Service_DefaultAPIKey if (Key is None) else Key,
                "prompt": {
                    "conversation": PromptConversation,
                    "parameters": PromptParameters
                },
                "user_parameters": UserParameters
            }
        }
        data["content"] = encryption.Encrypt(h, self.__server_public_key__, json.dumps(data["content"]), self.__configuration__.Encryption_Threads)
        await self.Send(json.dumps(data))

        redirectTo = None

        while (True):
            recvData = await self.Receive()
            recvData = json.loads(recvData)
            recvData = encryption.Decrypt(
                encryption.ParseHash(recvData["hash"]),
                self.__private_key__,
                recvData["data"],
                self.__configuration__.Encryption_Threads
            )
            token = json.loads(recvData)

            if ("redirect_to" in token):
                redirectTo = token["redirect_to"]
                break

            yield token

            if ("ended" in token and token["ended"]):
                break
        
        if (redirectTo is not None):
            previousConnection = None

            if (token["redirect_to"]["host"] is not None and token["redirect_to"]["port"] is not None):
                previousConnection = self.__current_connection__.copy()
                self.__socket_type__ = "websocket" if (token["redirect_to"]["type"] == "ws") else "socket" if (token["redirect_to"]["type"] == "s") else None

                await self.Connect(
                    Host = token["redirect_to"]["host"],
                    Port = token["redirect_to"]["port"],
                    Secure = token["redirect_to"]["secure"]
                )

            gen = self.AdvancedSendAndReceive(
                ModelName = token["redirect_to"]["model"],
                Key = Key,
                PromptConversation = PromptConversation,
                PromptParameters = PromptParameters,
                UserParameters = UserParameters,
                Service = Service
            )

            try:
                async for token in gen:
                    yield token
            except Exception as ex:
                raise ex
            finally:
                if (previousConnection is not None):
                    self.__socket_type__ = previousConnection[0]
                    await self.Connect(previousConnection[1], previousConnection[2], previousConnection[3])

                return
    
    async def GetAvailableModels(self, **kwargs) -> list[str]:
        models = None
        gen = self.AdvancedSendAndReceive("", Service = "get_available_models", **kwargs)

        async for token in gen:
            if ("models" in token):
                models = token["models"]
            
            if ("errors" in token and len(token["errors"]) > 0):
                raise RuntimeError(f"Unexpected server error(s): {token['errors']}")
        
        if (models is None):
            raise RuntimeError("Could not get models.")
        
        return models
    
    async def GetModelInfo(self, ModelName: str, **kwargs) -> dict[str, Any]:
        modelInfo = None
        gen = self.AdvancedSendAndReceive(ModelName, Service = "get_model_info", **kwargs)

        async for token in gen:
            if ("config" in token):
                modelInfo = token["config"]
            
            if ("errors" in token and len(token["errors"]) > 0):
                raise RuntimeError(f"Unexpected server error(s): {token['errors']}")
        
        if (modelInfo is None):
            raise RuntimeError("Could not get model information.")
        
        return modelInfo
    
    async def GetQueueData(self, ModelName: str, **kwargs) -> dict[str, int | float]:
        queueData = None
        gen = self.AdvancedSendAndReceive(ModelName, Service = "get_queue_data", **kwargs)

        async for token in gen:
            if ("queue" in token):
                queueData = token["queue"]
            
            if ("errors" in token and len(token["errors"]) > 0):
                raise RuntimeError(f"Unexpected server error(s): {token['errors']}")
        
        if (queueData is None):
            raise RuntimeError("Could not get queue data.")
        
        return queueData
    
    async def CreateAPIKey(
        self,
        Tokens: int = 0,
        ResetDaily: bool = False,
        ExpireDate: dict[str, int] | None = None,
        AllowedIPs: list[str] | None = None,
        PrioritizeModels: list[str] = [],
        Groups: list[str] = [],
        **kwargs
    ) -> str:
        key = None
        gen = self.AdvancedSendAndReceive("", PromptParameters = {
            "tokens": Tokens,
            "reset_daily": ResetDaily,
            "expire_date": ExpireDate,
            "allowed_ips": AllowedIPs,
            "prioritize_models": PrioritizeModels,
            "groups": Groups
        }, Service = "create_api_key", **kwargs)

        async for token in gen:
            if ("key" in token):
                key = token["key"]
            
            if ("errors" in token and len(token["errors"]) > 0):
                raise RuntimeError(f"Unexpected server error(s): {token['errors']}")

        if (key is None):
            raise RuntimeError("Could not create new API key.")
        
        return key
    
    async def DeleteAPIKey(self, Key: str, **kwargs) -> None:
        gen = self.AdvancedSendAndReceive("", PromptParameters = {
            "key": Key
        }, Service = "delete_api_key", **kwargs)

        async for token in gen:
            if ("errors" in token and len(token["errors"]) > 0):
                raise RuntimeError(f"Unexpected server error(s): {token['errors']}")
       
    async def GetKeyData(self, Key: str, **kwargs) -> dict[str, Any] | None:
        key = None
        gen = self.AdvancedSendAndReceive("", PromptParameters = {
            "key": Key
        }, Service = "get_key_data", **kwargs)

        async for token in gen:
            if ("key" in token):
                key = token["key"]
            
            if ("errors" in token and len(token["errors"]) > 0):
                raise RuntimeError(f"Unexpected server error(s): {token['errors']}")

        if (key is None):
            raise RuntimeError("Could not fetch key data.")
        
        return key
    
    async def BanUser(self, Type: Literal["key", "ip"], Value: str, **kwargs) -> None:
        gen = self.AdvancedSendAndReceive("", PromptParameters = {
            "type": Type,
            "value": Value
        }, Service = "ban", **kwargs)

        async for token in gen:
            if ("errors" in token and len(token["errors"]) > 0):
                raise RuntimeError(f"Unexpected server error(s): {token['errors']}")
    
    async def PardonUser(self, Type: Literal["key", "ip"], Value: str, **kwargs) -> None:
        gen = self.AdvancedSendAndReceive("", PromptParameters = {
            "type": Type,
            "value": Value
        }, Service = "pardon", **kwargs)

        async for token in gen:
            if ("errors" in token and len(token["errors"]) > 0):
                raise RuntimeError(f"Unexpected server error(s): {token['errors']}")
    
    async def GetSupport(self, **kwargs) -> list[dict[str, str]]:
        support = None
        gen = self.AdvancedSendAndReceive("", Service = "get_support", **kwargs)

        async for token in gen:
            if ("support" in token):
                support = token["support"]
            
            if ("errors" in token and len(token["errors"]) > 0):
                raise RuntimeError(f"Unexpected server error(s): {token['errors']}")

        if (support is None):
            raise RuntimeError("Could not fetch support data.")
        
        return support